#
# See the file "LICENSE" for the full license governing this code.
#
# frozen_string_literal: true

require 'easy_type'
require 'hocon/parser/config_document_factory'
require 'hocon/config_value_factory'
require 'fileutils'

# TODO: Add documentation
Puppet::Type.newtype(:fact_config) do
  include EasyType

  def self.module_name
    'easy_type'
  end

  desc <<-DESC
  Specify the configuration for the fact.

  blocked — Prevents all facts within the listed groups from being resolved when Facter runs.

  ttls — Caches the key-value pairs of groups and their duration to be cached.

  Caching and blocking facts is useful when Facter is taking a long time and slowing down your
  code. When a system has a lot of something — for example, mount points or disks — Facter
  can take a long time to collect the facts from each one. When this is a problem, you can
  speed up Facter’s collection by either blocking facts you’re uninterested in (blocklist),
  or caching ones you don’t need retrieved frequently (ttls).

  DESC

  def self.config_dir
    kernel = Facter.value(:kernel)
    case kernel
    when 'Linux'
      '/etc/puppetlabs/facter'
    when 'windows'
      'C:/ProgramData/PuppetLabs/facter/etc'      
    else
      fail "'#{kernel}' not (yet) supported."
    end
  end

  def self.config_file
    "#{config_dir}/facter.conf"
  end

  def self.config_data
    @config_data ||= Hocon::ConfigFactory.parse_file(config_file).resolve
  end

  def config_doc
    @config_doc ||= if File.exist?(self.class.config_file)
                      Hocon::Parser::ConfigDocumentFactory.parse_file(self.class.config_file)
                    else
                      Hocon::Parser::ConfigDocumentFactory.parse_string('')
                    end
  end

  attr_writer :config_doc

  def current_config
    @current_config ||= Hocon.load(self.class.config_file)['facts']
  end

  def write_config
    FileUtils.mkdir_p(self.class.config_dir) unless File.exists?(self.class.config_dir) 
    File.write(self.class.config_file, config_doc.render)
  end

  to_get_raw_resources do
    # Get all TTL's
    ttls = if config_data.has_path?('facts.ttls')
             Hash[config_data.get_value('facts.ttls').to_array.collect { |e| [e.keys.first, e.values.first.value] }]
           else
             {}
           end
    # Get the blocklist
    blocklist = if config_data.has_path?('facts.blocklist')
                  config_data.get_value('facts.blocklist').to_array.collect(&:value)
                else
                  []
                end
    # Merge lists of facts together
    all_facts = ttls.keys + blocklist
    # and fact for fact provide the information
    all_facts.collect do |fact|
      value = { :name => fact }
      value[:blocked] = blocklist.include?(fact)
      value[:ttl] = ttls[fact] if ttls.key?(fact)
      value
    end
  end

  on_modify do
    nil # Leave processing to propeties
  end

  after_modify do
    write_config
  end

  parameter :name
  property  :ttl
  property  :blocked
end
